import os
import json
import typing as T

import six
import yaml


class LoaderMeta(type):

    def __new__(mcs, __name__, __bases__, __dict__):  # type: ignore
        """Add include constructer to class."""

        # register the include constructor on the class
        cls = mcs.__class__.__new__(mcs, __name__, __bases__, __dict__)
        cls.add_constructor('!include', cls.construct_include)

        return cls


@six.add_metaclass(LoaderMeta)
class Loader(yaml.Loader):
    """YAML Loader with `!include` constructor."""

    def __init__(self, stream, filenames_seen=None):
        # type: (T.IO[str], T.Optional[T.Set[str]]) -> None
        """Initialise Loader."""

        try:
            self._root = os.path.split(stream.name)[0]
        except AttributeError:
            self._root = os.path.curdir

        self._filenames_seen = filenames_seen or set()
        yaml.Loader.__init__(self, stream)

    def _remeber(self):
        # type: () -> T.Callable[[T.IO[str]], Loader]
        """Create a loader with the same filenames cache."""
        def loader(stream):
            # type: (T.IO[str]) -> Loader
            return Loader(stream, self._filenames_seen)
        return loader

    def construct_include(self, node):
        # type: (T.Any) -> T.Any
        """Include file referenced at node."""

        filename = os.path.abspath(
            os.path.join(
                self._root, self.construct_scalar(node)))  # type: ignore
        extension = os.path.splitext(filename)[1].lstrip('.')

        if filename in self._filenames_seen:
            raise ValueError('Circular !include {}'.format(filename))
        self._filenames_seen.add(filename)

        with open(filename, 'r') as f:
            if extension in ('yaml', 'yml'):
                return yaml.load(f, self._remeber())
            if extension == 'json':
                return json.load(f)
            return f.read()



from typing import Any, Callable, IO, Iterator, Set, Tuple, Union

Obj = Any
Path = Tuple[Any, ...]
WalkType = Iterator[Tuple[Path, Obj]]
Cache = Set[int]
ObjI = Iterator[Tuple[Any, Any]]
ObjIFunc = Callable[[Obj], ObjI]
Stream = Union[str, IO[str]]



import tempfile
import unittest

import yamlenv


class TestYamlInclude(unittest.TestCase):

    def test_non_existing(self):
        with self.assertRaises(IOError):
            yamlenv.load('''
    a: 1
    b: !include non-existing.yaml
            ''')

    def test_empty_file(self):
        f = tempfile.NamedTemporaryFile(suffix='.yaml')

        data = yamlenv.load('''
a: 1
b: !include {}
        '''.format(f.name))

        self.assertEqual(data, {'a': 1, 'b': None})

    def test_include(self):
        f = tempfile.NamedTemporaryFile(suffix='.yaml')
        f.write(b'2')
        f.flush()

        data = yamlenv.load('''
a: 1
b: !include {}
        '''.format(f.name))

        self.assertEqual(data, {'a': 1, 'b': 2})

    def test_circular_self(self):
        f = tempfile.NamedTemporaryFile(suffix='.yaml')

        f.write('''
a: !include {}
        '''.format(f.name).encode())
        f.flush()
        f.seek(0)

        with self.assertRaises(ValueError):
            yamlenv.load(f)

    def test_circular(self):
        f1 = tempfile.NamedTemporaryFile(suffix='.yaml')
        f2 = tempfile.NamedTemporaryFile(suffix='.yaml')

        f1.write('''
a: !include {}
        '''.format(f2.name).encode())
        f1.flush()
        f1.seek(0)

        f2.write('''
a: !include {}
        '''.format(f1.name).encode())
        f2.flush()
        f2.seek(0)

        with self.assertRaises(ValueError):
            yamlenv.load(f1)



import os
import re
import yaml
import yaml.parser

try:
    from collections.abc import Mapping, Sequence, Set
except ImportError:
    # Python 2.7
    from collections import Mapping, Sequence, Set  # noqa

import six
import typing as T

from yamlenv.types import Cache, Obj, ObjIFunc, Path, WalkType


def objwalk(obj, path=None, memo=None):
    # type: (Obj, T.Optional[Path], T.Optional[Cache]) -> WalkType
    if path is None:
        path = tuple()
    if memo is None:
        memo = set()
    iterator = None  # type: T.Optional[ObjIFunc]
    if isinstance(obj, Mapping):
        iterator = six.iteritems
    elif isinstance(
            obj, (Sequence, Set)
    ) and not isinstance(obj, six.string_types):
        iterator = enumerate
    if iterator:
        if id(obj) not in memo:
            memo.add(id(obj))
            for path_component, value in iterator(obj):
                for result in objwalk(value, path + (path_component,), memo):
                    yield result
            memo.remove(id(obj))
    else:
        yield path, obj


class EnvVar(object):
    '''Follow Bash expansion rules.
    https://www.gnu.org/software/bash/manual/html_node\
/Shell-Parameter-Expansion.html
    '''
    __slots__ = ['name', 'separator', 'default', 'string']

    RE = re.compile(
        r'\$\{(?P<name>[^:-]+?)((?P<separator>:?)-(?P<default>.*?))?\}')

    def __init__(self, name, separator, default, string):
        # type: (str, str, str, str) -> None
        self.name = name
        self.separator = separator
        self.default = default
        self.string = string

    @property
    def allow_null_default(self):
        # type: () -> bool
        return self.separator == ''

    @property
    def value(self):
        # type: () -> str
        value = os.environ.get(self.name)
        if value:
            return self.RE.sub(value, self.string, count=1)
        if self.allow_null_default or self.default:
            return self.RE.sub(self.default, self.string, count=1)
        raise ValueError('Missing value and default for {}'.format(self.name))

    @property
    def maybe_yaml_value(self):
        # type: () -> T.Any
        v = self.value
        try:
            return yaml.safe_load(v)
        except yaml.parser.ParserError:
            return v

    @classmethod
    def from_string(cls, s):
        # type: (str) -> T.Optional[EnvVar]
        if not isinstance(s, six.string_types):
            return None
        data = cls.RE.search(s)
        if not data:
            return None
        gd = data.groupdict()
        return cls(gd['name'], gd['separator'], gd['default'], s)


def interpolate(data):
    # type: (T.Any) -> Obj
    while True:
        more = False
        for path, obj in objwalk(data):
            e = EnvVar.from_string(obj)
            if e is not None:
                more = True
                x = data
                for k in path[:-1]:
                    x = x[k]
                x[path[-1]] = e.maybe_yaml_value
        if not more:
            break
    return data



import os
import unittest

import yamlenv


class TestYamlEnv(unittest.TestCase):

    def test_load(self):
        self.assertEqual(yamlenv.load('''
a: 1
b: 2
        '''), {'a': 1, 'b': 2})

    def test_load_all(self):
        self.assertEqual(list(yamlenv.load_all('''
a: 1
b: 2
---
c: 3
        ''')), [{'a': 1, 'b': 2}, {'c': 3}])

    def test_fail_if_no_env(self):
        with self.assertRaises(ValueError):
            yamlenv.load('''
a: ${A}
b: 2
            ''')

    def test_interpolate_integer(self):
        os.environ['A'] = '1'
        self.assertEqual(yamlenv.load('''
a: ${A}
b: 2
            '''), {'a': 1, 'b': 2})

    def test_interpolate_string(self):
        os.environ['A'] = 'password'
        self.assertEqual(yamlenv.load('''
a: ${A}
b: 2
        '''), {'a': 'password', 'b': 2})

    def test_interpolate_two_values(self):
        os.environ['A'] = '1'
        os.environ['B'] = 'foo'
        self.assertEqual(yamlenv.load('''
a: ${A} ${B}
            '''), {'a': "1 foo"})

    def test_interpolate_two_values_with_defaults(self):
        self.assertEqual(yamlenv.load('''
a: ${C:-foo} ${D:-bar}
            '''), {'a': "foo bar"})

    def test_interpolate_invalid_yaml_value(self):
        self.assertEqual(yamlenv.load('''
{'a': {'b': '{foo} ${C:-foo}'}}
            '''), {'a': {'b': "{foo} foo"}})

    def test_interpolate_within_characters(self):
        os.environ['A'] = 'def'
        self.assertEqual(yamlenv.load('''
a: abc${A}ghi
b: 2
        '''), {'a': 'abcdefghi', 'b': 2})

    def test_interpolate_string_alternative_separator(self):
        os.environ['A'] = 'password'
        self.assertEqual(yamlenv.load('''
a: ${A:-1}
b: 2
        '''), {'a': 'password', 'b': 2})

    def test_interpolate_string_prefer_env(self):
        os.environ['A'] = 'password'
        self.assertEqual(yamlenv.load('''
a: ${A-1}
b: 2
        '''), {'a': 'password', 'b': 2})

    def test_interpolate_default(self):
        self.assertEqual(yamlenv.load('''
a: ${A-1}
b: 2
            '''), {'a': 1, 'b': 2})

    def test_interpolate_default_alternative_separator(self):
        self.assertEqual(yamlenv.load('''
a: ${A:-1}
b: 2
            '''), {'a': 1, 'b': 2})

    def test_interpolate_nested(self):
        self.assertEqual(yamlenv.load('''
a: 1
b:
  b1: 21
  b2: ${B-22}
            '''), {'a': 1, 'b': {'b1': 21, 'b2': 22}})

    def test_interpolate_embedded(self):
        self.assertEqual(yamlenv.load('''
a: ${FOO:-foo} bar
'''), {'a': 'foo bar'})

    def test_interpolate_embedded_integer(self):
        self.assertEqual(yamlenv.load('''
a: ${FOO:-7} bar
'''), {'a': '7 bar'})


    def test_interpolate_default_empty(self):
        self.assertEqual(yamlenv.load('''
a: bar ${FOO-} bar
'''), {'a': 'bar  bar'})

    def test_default_empty(self):
        self.assertEqual(yamlenv.load('''
a: ${FOO-}
'''), {'a': None})

    def test_default_does_not_exist(self):
        with self.assertRaises(ValueError):
            yamlenv.load('''
a: ${FOO:-} bar
''')



import typing as T
import yaml

from yamlenv import env, loader
from yamlenv.types import Stream


__version__ = '0.7.1'


def load(stream):
    # type: (Stream) -> T.Any
    data = yaml.load(stream, loader.Loader)
    return env.interpolate(data)


def load_all(stream):
    # type: (Stream) -> T.Iterator[T.Any]
    for data in yaml.load_all(stream, loader.Loader):
        yield env.interpolate(data)



from setuptools import setup, find_packages

with open('README.rst') as readme:
    long_description = readme.read()

setup(
    name='yamlenv',
    version='0.7.1',
    author="Lorenzo Bolla",
    author_email="lbolla@gmail.com",
    description="Interpolate Yaml files with env vars",
    long_description=long_description,
    url="https://github.com/lbolla/yamlenv",
    package_data={'yamlenv': ['py.typed']},
    packages=find_packages('.'),
    install_requires=[
        'PyYAML>=3.12',
        'six>=1.10',
        'typing; python_version<"3.6"',
    ],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 3",
    ],
)
